/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
/* Thonis is the app singleton class */

package net.hoopajoo.cheeselab3;

import javax.imageio.*;
import javax.imageio.stream.*;
import javax.swing.*;        

import net.hoopajoo.swingx.*;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

public class CheeseMap {
	static CheeseProgress progress = null;
	static String cachedir = "cache/_null";
	static int edit_roomid = 0;	
	static JFrame toplevel = null;
	static Socket sock = null;
	static BufferedInputStream in = null;
	static PrintStream out = null;
	static CheeseCanvas canvas = null;
	static CheeseMapData map = null;
	static CheeseMapRoomEditor room_editor = null;
	static CheeseMapNavigator navi = null;
	static CheeseMapSearch find = null;
	static CheeseMapFollowDialog select_follow = null;
	static JToolBar toolbar = null;
	static CheeseMapToolbox toolbox = null;
	static volatile boolean online = false;
	
	
	static SelectedAction connect_action;
	static SelectedAction follow_action;
	static SelectedAction zoomin_action;
	static SelectedAction zoomout_action;
	static SelectedAction levelup_action;
	static SelectedAction leveldown_action;
	static SelectedAction find_action;
	static SelectedAction navi_action;
	
	static String follow_userid = "";
	static boolean follow_player = true;
	
	static volatile boolean pinger = false;
	
	static Properties props = null;
	
	static NetworkBonanza bozo = null;
	static int last_acceptid, last_sendid;

	/*
	public static void die( String s ) {
		dmesg( s );
	}

	public static void dmesg( String s ) {
		System.out.println( s );
	}
	*/

	private static void createAndShowGUI() {
		//JFrame.setDefaultLookAndFeelDecorated( true );
		JFrame.setDefaultLookAndFeelDecorated( false );
		toplevel = new JFrame( "CheeseMap v" + CheeseLab.version );
		toplevel.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

		Container content = toplevel.getContentPane();

		// Canvas
		canvas = new CheeseCanvas();

		// Create actions
		connect_action = new SelectedAction() {
		    public void actionPerformed(ActionEvent e) {
		        String cmd = e.getActionCommand();
		        connect_action.setSelected( !isSelected() );
		        if( cmd.equals( "connect" ) ) {
					CheeseLab.dmesg( 3, "Connecting" );
					connect();
				}else if( cmd.equals( "disconnect" )) {
					CheeseLab.dmesg( 3, "Disconnecting" );
					disconnect();
				}
		    }
		};
		
		follow_action = new SelectedAction() {
		    public void actionPerformed(ActionEvent e) {
		        set_follow_player( !isSelected() );
		    }
		};
		follow_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Follow Player" );
		follow_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( 'f' ) );
		
		zoomin_action = new SelectedAction() {
			public void actionPerformed( ActionEvent e ) {
				CheeseMap.zoom_in();
			}
		};
		zoomin_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Zoom IN (+)" );
		zoomin_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( '+' ) );
		
		zoomout_action = new SelectedAction() {
			public void actionPerformed( ActionEvent e ) {
				CheeseMap.zoom_out();
			}
		};
		zoomout_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Zoom OUT (-)" );
		zoomout_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( '-' ) );
		
		levelup_action = new SelectedAction() {
		    public void actionPerformed( ActionEvent e ) {
		        CheeseMap.level_up();
		    }
		};
		levelup_action.putValue( SelectedAction.SHORT_DESCRIPTION, "UP one level" );
		levelup_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( KeyEvent.VK_UP,
		                				InputEvent.CTRL_MASK ) );

		leveldown_action = new SelectedAction() {
		    public void actionPerformed( ActionEvent e ) {
		        CheeseMap.level_down();
		    }
		};
		leveldown_action.putValue( SelectedAction.SHORT_DESCRIPTION, "DOWN one level" );
		leveldown_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( KeyEvent.VK_DOWN,
		                				InputEvent.CTRL_MASK ) );
		
		find_action = new SelectedAction() {
			public void actionPerformed( ActionEvent e ) {
			    CheeseMap.find.setVisible( true );
			}
		};
		find_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Find" );
		find_action.putValue( SelectedAction.ACCELERATOR_KEY,
		        KeyStroke.getKeyStroke( KeyEvent.VK_F,
		                				InputEvent.CTRL_MASK ) );
		
		navi_action = new SelectedAction() {
		    public void actionPerformed( ActionEvent e ) {
		        setSelected( !isSelected() );
		        CheeseMap.navi.setVisible( isSelected() );
		    }
		};
		navi_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Navigator" );
				
		// Create toolbox 
		toolbox = new CheeseMapToolbox();
		
		// Create toolbar
		toolbar = new JToolBar();
		
		// Setup menus
		JMenuBar bar = new JMenuBar();
		
		// File menu
		JMenu menu = new JMenu( "File" );
		menu.setMnemonic( KeyEvent.VK_F );
		JMenuItem item;
		
		menu.add( new JSelectedMenuItem( connect_action ) );
		
		item = new JMenuItem( "Preferences", 'P' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.show_preferences();
		    }
		} );
		menu.add( item );
		item = new JMenuItem( "Change Follow", 'F' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.select_follow.setVisible( true );
		    }
		} );
		menu.add( item );
		item = new JMenuItem( "Change Player Icon", 'I' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.change_avatar();
		    }
		} );
		menu.add( item );
		item = new JMenuItem( "Sync Cache", 'H' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.do_refresh();
		    }
		} );
		menu.add( item );
		menu.addSeparator();
		item = new JMenuItem( "Exit", 'X' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.exit();
		    }
		} );
		menu.add( item );
		bar.add( menu );
		
		// Editor menu
		menu = new JMenu( "Edit" );
		menu.setMnemonic( 'E' );
		menu.add( new JSelectedMenuItem( CheeseMap.find_action ) );
		menu.add( toolbox.createMenu( "Tools" ) );
		menu.addSeparator();
		item = new JMenuItem( "Merge All", 'A' );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JOptionPane.showMessageDialog( toplevel, "Not done, sorry" );
			}
		} );
		menu.add( item );
		bar.add( menu );
		
		// Editor menu
		menu = new JMenu( "View" );
		menu.setMnemonic( 'V' );
		JMenu submenu = new JMenu( "Toolbars" );
		item = new JCheckBoxMenuItem( "Navigation Toolbar", toolbar.isVisible() );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    toggleToolbar( toolbar,
			            ((JCheckBoxMenuItem)e.getSource()).isSelected() );
			}
		} );
		submenu.add( item );
		item = new JCheckBoxMenuItem( "Editor Toolbar", toolbox.isVisible() );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    toggleToolbar( toolbox,
			            ((JCheckBoxMenuItem)e.getSource()).isSelected() );
			}
		} );
		submenu.add( item );
		menu.add( submenu );
		menu.add( new JSelectedCheckMenuItem( navi_action ) );
		menu.add( new JSelectedCheckMenuItem( follow_action ) );
		menu.add( new JSelectedMenuItem( levelup_action ) );
		menu.add( new JSelectedMenuItem( leveldown_action ) );
		menu.add( new JSelectedMenuItem( zoomin_action ) );
		menu.add( new JSelectedMenuItem( zoomout_action ) );

		item = new JCheckBoxMenuItem( "Render Notes", canvas.render_notes );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.canvas.render_notes = ((JCheckBoxMenuItem)e.getSource()).isSelected();
				CheeseMap.canvas.repaint();
			}
		} );
		menu.add( item );
		
		item = new JCheckBoxMenuItem( "Render Futures", canvas.render_future );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.canvas.render_future = ((JCheckBoxMenuItem)e.getSource()).isSelected();
				CheeseMap.canvas.repaint();
			}
		} );
		menu.add( item );

		item = new JCheckBoxMenuItem( "Render Grid", canvas.render_grid );
		item.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				CheeseMap.canvas.render_grid = ((JCheckBoxMenuItem)e.getSource()).isSelected();
				CheeseMap.canvas.set_tile_size( CheeseMap.canvas.tile_width, CheeseMap.canvas.tile_height ); // this will clear the cache
				CheeseMap.canvas.repaint();
			}
		} );
		menu.add( item );

		bar.add( menu );

		toplevel.setJMenuBar( bar );
		
		// Setup toolbar
		JButton bb;
		JSelectedButton sb;
		
		// "Generic" commands (connect, follow etc.)
		sb = new JSelectedButton( connect_action );
		sb.setIcon( new ImageIcon( "ui/ui_disconnect.png" ) );
		sb.setSelectedIcon( new ImageIcon( "ui/ui_connect.png" ) );
		toolbar.add( sb );
		set_connect_status( false );
		
		bb = new JButton( find_action ); 
		bb.setIcon( new ImageIcon( "ui/ui_find.png" ) );
		toolbar.add( bb );

		sb = new JSelectedButton( follow_action );
		sb.setIcon( new ImageIcon( "ui/ui_follow.png" ) );
		toolbar.add( sb );
		
		// Zooming
		toolbar.addSeparator();
		bb = new JButton( zoomin_action );
		bb.setIcon( new ImageIcon( "ui/ui_zoomin.png" ) );
		toolbar.add( bb );
		bb = new JButton( zoomout_action ); 
		bb.setIcon( new ImageIcon( "ui/ui_zoomout.png" ) );
		toolbar.add( bb );
		
		// Navigation
		toolbar.addSeparator();
		bb = new JButton( levelup_action );
		bb.setIcon( new ImageIcon ( "ui/ui_u.png" ) );
		toolbar.add ( bb );
		
		bb = new JButton( leveldown_action );
		bb.setIcon( new ImageIcon ( "ui/ui_d.png" ) );
		toolbar.add ( bb );
		
		// Add toolbars to main window
		content.setLayout(new AKDockLayout());
		toggleToolbar( toolbar, true );
		toggleToolbar( toolbox, true );

		map = new CheeseMapData();

		content.add( canvas, AKDockLayout.CENTER );
		canvas.set_map( map );

		toplevel.pack();
		// TODO: remember last size/position
		Toolkit tk = Toolkit.getDefaultToolkit();
		Dimension dm = tk.getScreenSize();
		toplevel.setSize( (int)dm.getWidth(), (int)dm.getHeight() / 3 );
		toplevel.setVisible( true );
		
		load_properties();
		set_connect_status ( false );
		set_follow_player( true );
		
		// load the room editor
		room_editor = new CheeseMapRoomEditor( toplevel );
		find = new CheeseMapSearch( toplevel );
		navi = new CheeseMapNavigator( toplevel );
		progress = new CheeseProgress( toplevel );
		select_follow = new CheeseMapFollowDialog( toplevel );
	}

	public static void set_edit_room( int roomid ) {
	    set_edit_room( roomid, true );
	}
	
	public static void set_edit_room( CheeseRoomData r ) {
	    set_edit_room( r, true );
	}
	
	public static void set_edit_room( int roomid, boolean open_editor ) {
		CheeseLab.dmesg( 2, "setting edit room# to " + roomid );
		// Make sure it gets in the cache
		map.get_room( roomid );
		
		// TODO: make sure we get the latest?
		CheeseMap.clearping( "set_edit_room" );
		send_tx( 1, "ping", "set_edit_room" );
		while( ! CheeseMap.checkping() );

		set_edit_room( map.get_room( roomid ), open_editor );
	}
	
	public static void set_edit_room( CheeseRoomData r, boolean open_editor ) {
		edit_roomid = r.roomid;
		canvas.center_on( r.x, r.y, r.z );
		room_editor.set_room( r );
		set_follow_player( false );
		if( open_editor )
		    room_editor.setVisible( true );
	}
	
	public static void save_edit_room() {
		CheeseRoomData r = room_editor.get_room();
		send_tx( 1, "save", CheeseDecoder.encode_object( r ) );
		canvas.changed();
	}

	private static void show_preferences() {
		CheeseMapConfig c = new CheeseMapConfig( toplevel, props );
		c.setVisible( true );
	}
	
	private static void load_properties() {
		props = new Properties();
		try {
			FileInputStream in = new FileInputStream( "cheesemap.cfg" );
			props.load( in );
			//props.list( System.out );
			in.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}
		CheeseMap.apply_properties();
	}
	
	public static void apply_properties() {
		// apply properties
		follow_userid = props.getProperty( "userid" );
		try {
			int loglevel = Integer.parseInt( props.getProperty( "loglevel" ) );
			CheeseLab.log_level = loglevel;
		}catch( Exception e ) {
			CheeseLab.log_level = 0;
		}
	}
	
	public static void save_properties() {
		try {
			FileOutputStream out = new FileOutputStream( "cheesemap.cfg" );
			props.store( out, null );
			out.close();
		}catch( Exception e ) {
			e.printStackTrace();
		}		
	}
	
	/**
	 * Set connection status of components that provide visual feed-back to user
	 * @param connected  true if mapper is in connected state, false othervise.
	 */
	private static void set_connect_status(boolean connected) {
		if (connected) {
		    connect_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Disconnect" );
		    connect_action.putValue( SelectedAction.ACTION_COMMAND_KEY, "disconnect" );
		    connect_action.putValue( SelectedAction.MNEMONIC_KEY,
		            new Integer( KeyEvent.VK_D ) );
			connect_action.setSelected( true );
		}
		else {
		    connect_action.putValue( SelectedAction.SHORT_DESCRIPTION, "Connect" );
		    connect_action.putValue( SelectedAction.ACTION_COMMAND_KEY, "connect" );
		    connect_action.putValue( SelectedAction.MNEMONIC_KEY,
		            new Integer( KeyEvent.VK_C ) );
			connect_action.setSelected( false );
		}
	}
	
	private static void toggleToolbar( JToolBar bar, boolean visible ) {
	    Container content = toplevel.getContentPane();
	    if ( visible ) {
	        bar.setVisible( true );
	        bar.setEnabled( true );
	        content.add( bar, AKDockLayout.NORTH );
	        content.validate();
	    } else {
	        // FIXME: If toolbar is undocked this beahves rather strangely
	        Container parent = bar.getParent();
	        bar.setVisible( false );
	        bar.setEnabled( false );
	        parent.remove( bar );
	        parent.validate();
	    }
	}
	
	public static void zoom_in() {
		resize_tiles( 5 );
	}
	
	public static void zoom_out() {
		resize_tiles( -5 );
	}
	
	public static void level_up() {
		CheeseMap.set_follow_player( false );
		canvas.move_origin( 0, 0, 1 );
	}
	
	public static void level_down() {
		set_follow_player( false );
		canvas.move_origin( 0, 0, -1 );
	}
	
	private static void disconnect() {
		CheeseLab.dmesg( 2, "disconnect" );
		
		set_connect_status( false );
		
		if( sock != null ) {

			if( bozo != null ) bozo.running = false;

			try { in.close();	}catch( Exception e ) {	}
			try { out.close();	}catch( Exception e ) {	}
			try { sock.close();	}catch( Exception e ) {	}

			in = null;
			out = null;
			sock = null;
			bozo = null;
		}
		
		online = false;
		canvas.repaint();
	}

	private static void connect() {
		disconnect();

		// Make sure username is right
		apply_properties();
		
		String password = "";
		
		if( props.getProperty( "password" ).equals( "PROMPT" ) )
			password = JOptionPane.showInputDialog( toplevel, "Password" );
		else
			password = props.getProperty( "password" );
		
		try {
			sock = new Socket( props.getProperty( "server" ), Integer.parseInt( props.getProperty( "port" ) ) );
			in = new BufferedInputStream( sock.getInputStream() );
			out = new PrintStream( sock.getOutputStream(), true );
			bozo = new NetworkBonanza();
			
			CheeseLab.dmesg( 1, "Authenticating" );
			send_tx( 1, "auth", follow_userid + ":" + password );

			set_connect_status ( true );
			
			cachedir = "cache/" +  props.getProperty( "server" ) + "_" + props.getProperty( "port" );
			map.initcache();
			
			// Send our player tile
			send_player_tile();
			set_follow_player( true );
		}catch( Exception e ) {
			CheeseLab.dmesg( 0, "Unable to connect: " + e );
			JOptionPane.showMessageDialog( toplevel, "Unable to connect: " + e );
			set_connect_status ( false );
		}
	}

	public static void send_player_tile() {
		CheeseFile f = new CheeseFile( "tiles/player_" + follow_userid + ".png" );
		send_tx( 1, "tile", CheeseDecoder.encode_object( f ) );
	}
	
	public static void do_refresh() {
		online = false; // tells map not to draw
		if( JOptionPane.showConfirmDialog( toplevel, "Are you sure you want to sync? If the database is large and you have not sync'd before it may take a loooong time!", "Confirm Sync", JOptionPane.YES_NO_OPTION ) == JOptionPane.YES_OPTION ) {			
			CheeseLab.dmesg( 2, "Performing pull of all new stuff: " + map.get_highest_mtime() );
			// Refresh should be mtime of last newest item in our cache
			send_tx( 1, "refresh", Long.toString( map.get_highest_mtime() ) );
		}else{
			online = true;
		}
	}
	
	public static int send_tx( int channel, String msgtype, String data ) {
		return( send_tx( channel, msgtype, data.getBytes() ) );
	}
	
	public static int send_tx( int channel, String msgtype, byte[] data ) {
		if( sock == null )
			return( -1 );

		last_sendid = CheeseTx.send_tx( out, last_sendid, channel, msgtype, data );
		return( 0 );		
	}

	public static int handle_tx() {
		if( sock == null )
			return( -1 );

		CheeseLab.dmesg( 2, "Handle tx" );
		CheeseTx tx = CheeseTx.accept_tx( in );
		try {
			
			// Just pull the stuff from the tx as we did it manually before
			// TODO: make this nice
			if( tx == null ) {
				CheeseLab.dmesg( 2, "Got null tx, bye" );
				disconnect();
				return( -1 );
			}
			
			int seq = tx.sequence;
			int channel = tx.channel;
			String msgtype = tx.msgtype;
			String enc = tx.enc;
			int len = tx.length;

			byte[] buffer = tx.buffer;

			CheeseLab.dmesg( 2, "msgtype is: " + msgtype );
			
			// Now based on msttype, pass the data to the proper routine
			if( msgtype.equals( "player" ) ) {
				map.update_player( buffer );
				// Set canvas x/y/z (center)
				CheesePlayerData p = map.get_player( follow_userid );

				if( p != null && follow_player ) {
					CheeseLab.dmesg( 3, "Moving origin: " + p.x + "," + p.y + "," + p.z );
					canvas.center_on( p.x, p.y, p.z );
				}
			}else if( msgtype.equals( "progress" ) ) {
				// Display a progress message, if blank then no message to display
				String message = new String( tx.buffer );
				//CheeseLab.dmesg( 0, message );
				
				if( message.equals( "CLOSE" ) ) {
					progress.setVisible( false );
				}else{
					progress.set_message( message );
					if( ! progress.isShowing() ) 
						progress.setVisible( true );
				}
			}else if( msgtype.equals( "search" ) ) {
				// Search results
				find.set_search_results( buffer );
			}else if( msgtype.equals( "auth" ) ) {
				// Now we are connected and can do stuff
				online = true;
			}else if( msgtype.equals( "ping" ) ) {
				// pings allow network sync, set pingmonitor(message)=0
				// then wait for pingmonitor(message)=1, since
				// network ops happen in sequence your response will be complete
				// when you receieve your ping
				//setping( new String( buffer ) );
				setping();
			}else if( msgtype.equals( "tile" ) ) {
				// Accept tile, the format is tilename,tiledata
				CheeseFile f = (CheeseFile)CheeseDecoder.decode_object( tx.buffer );
				f.save( cachedir + "/tiles/" );
				CheeseLab.dmesg( 2, "Accepting tile: " + f.filename );
				
				// Invalidate map cache (we should just invalidate this tile, but oh well)
				canvas.set_tile_size( canvas.tile_width, canvas.tile_height );
			}else if( msgtype.equals( "message" ) ) {
				// Display message to user
				// TODO: need a chat thing too? :p
				CheeseLab.dmesg( 2, "Got message: " + new String( tx.buffer ) );
				JOptionPane.showMessageDialog( toplevel, new String( tx.buffer ) );
			}else if( msgtype.equals( "refresh" ) ) {
				online = true;
				// refresh complete
				map.save_max_mtime();
			}else if( msgtype.equals( "map" ) ) {
				canvas.update_map( buffer );
			}else if( msgtype.equals( "world" ) ) {
				// TODO: use this in drawing optimization
				canvas.changed();
			}else if( msgtype.equals( "room" ) ) {
				map.update_room( buffer );
				canvas.changed();
			}else{
				CheeseLab.dmesg( 2, "Unknown message type: " + msgtype );
			}
		}catch( Exception e ) {
			disconnect();
		}
		return( 0 );
	}

	public static void setping() {
		pinger = true;
	}
	
	// WARNING: due to syncro issues there is only 1 pinger now
	public static boolean checkping() {
		if( sock == null )
			return( true );

		return( pinger );
	}

	public static void clearping( String msg ) {
		pinger = false;
	}

	public static void set_edit_tool( String toolname ) {
		if (toolname.equals("move_tool")) {
			JOptionPane.showMessageDialog( toplevel, "Not done, sorry!\nUse keyboard to move rooms around for now." );
			toolbox.setActiveTool( "select_tool" );
		}
		else
			canvas.setEditTool (toolname);
	}

	public static void set_follow_player( boolean f ) {
	    follow_action.setSelected( f );
		follow_player = f;
		
		// Get the playername and re-center on it if we can
		CheesePlayerData p = map.get_player( follow_userid );
		if( p != null && follow_player ) {
			canvas.center_on( p.x, p.y, p.z );
		}
	}
	
	public static void main(String[] args) {
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
	    
		// Make sure that windows behave "natively" from Java 1.5 onwards
		System.setProperty( "java.awt.Window.locationByPlatform", "true" );
		
		javax.swing.SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				createAndShowGUI();
				CheeseLab.dmesg( 2, "End of main_run()" );
			}
		} );
	}
	
	public static void resize_tiles( int howmuch ) {
		set_tile_size( canvas.tile_width + howmuch );
	}

	public static void set_tile_size( int size ) {
		canvas.set_tile_size( size, size );
	}
	
	public static void change_avatar() {
		JFileChooser chooser = new JFileChooser();
		
		chooser.showDialog( toplevel, "Select Avatar Image" );
		File f = chooser.getSelectedFile();
		
		if( f != null ) {
			// Read the image, and save it out.  If we are online, also push it to the server
			CheeseLab.dmesg( 0, "They chose: " + f );
			ImageIcon ii = new ImageIcon( f.getAbsolutePath() );
			
			if( ii.getIconWidth() == -1 ) {
				// Error
				JOptionPane.showMessageDialog( toplevel, "Error loading image" );
			}else{
				Image i = ii.getImage().getScaledInstance( 128, 128, Image.SCALE_SMOOTH );
				
				// Force loading
				while( ! canvas.prepareImage( i, -1, -1, (ImageObserver)canvas ) );
	
				// Resize
				BufferedImage im = new BufferedImage( 128, 128, Transparency.TRANSLUCENT );
				Graphics2D g = im.createGraphics();
				g.drawImage( i, 0, 0, null );
				g.dispose();
				
				Iterator writers = ImageIO.getImageWritersByFormatName( "png" );
				ImageWriter writer = (ImageWriter)writers.next();
	
				File output = new File( "tiles/player_" + follow_userid + ".png" );
				try {
					ImageOutputStream ios = ImageIO.createImageOutputStream( output );
					writer.setOutput( ios );
					writer.write( im );
					
					ios.close();
					
					// Clear tiles
					canvas.set_tile_size( canvas.tile_width, canvas.tile_height );
					if( sock != null )
						send_player_tile();
					
					JOptionPane.showMessageDialog( toplevel, "Avatar changed" );
				}catch( Exception e ) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * Exit cleanly
	 */
	public static void exit() {
		disconnect();
		System.exit( 0 );
	}
}
